@@ -4,12 +4,12 @@ use std::{collections::BTreeMap, iter::zip};
 use rr_util::{
     symbolic_size::{SymbolicSizeConstraint, SymbolicSizeConstraints},
     tensor_util::{Shape, TorchDeviceDtypeOp},
+    IndexSet,
 };
 use uuid::uuid;
 
 use crate::{
-    CachedCircuitInfo, CircuitFlags, CircuitNode, CircuitNodeAutoName, ConstructError, NamedAxes,
-    Result,
+    CachedCircuitInfo, CircuitFlags, CircuitNode, ConstructError, NamedAxes, Result, Symbol,
 };
 
 pub trait CircuitNodeInit {
@@ -99,6 +99,21 @@ pub trait CircuitNodeComputeInfoImpl: CircuitNode {
         }
         result
     }
+
+    fn compute_free_symbols(&self) -> IndexSet<Symbol> {
+        let mut out = IndexSet::default();
+        for child in self.children() {
+            // ideally, symbol would just return itself rather than doing this check here
+            // but, then the info on that symbol wouldn't be initialized properly because this is called before init finishes
+            // So we can only check if children are initialize and then we special case syms in various places. : /
+            if let Some(sym) = child.as_symbol() {
+                out.insert(sym.clone());
+            } else {
+                out.extend(child.info().free_symbols.iter().cloned())
+            }
+        }
+        out
+    }
 }
 
 pub trait CircuitNodeSetNonHashInfo: CircuitNodePrivate {
@@ -112,6 +127,7 @@ impl<T: CircuitNodeComputeInfoImpl + CircuitNodePrivate> CircuitNodeSetNonHashIn
         self.info_mut().device_dtype = self.compute_device_dtype()?;
         self.info_mut().symbolic_size_constraints = self.compute_symbolic_size_constraints()?;
         self.info_mut().named_axes = self.compute_named_axes();
+        self.info_mut().free_symbols = self.compute_free_symbols();
         Ok(())
     }
 }
@@ -133,29 +149,12 @@ impl<T: CircuitNodeHashItems + CircuitNode> CircuitNodeHashWithChildren for T {
     }
 }
 
-pub trait CircuitNodeMaybeAutoName {
-    fn maybe_auto_name(&self, name: Option<String>) -> Option<String> {
-        name
-    }
-}
-impl<T: CircuitNodeAutoName> CircuitNodeMaybeAutoName for T {
-    fn maybe_auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            self.info()
-                .use_autoname()
-                .then(|| self.auto_name())
-                .flatten()
-        })
-    }
-}
-
 impl<T> CircuitNodeInit for T
 where
     T: CircuitNodeHashWithChildren
         + CircuitNodePrivate
         + CircuitNodeSetNonHashInfo
         + CircuitNode
-        + CircuitNodeMaybeAutoName
         + Sized,
 {
     fn init_info_impl(mut self, is_initial: bool) -> Result<Self> {
@@ -178,7 +177,13 @@ where
         }
         // we have to call maybe_auto_name after setting the flag
         if is_initial {
-            *self.name_mut() = self.maybe_auto_name(self.name_cloned());
+            *self.name_mut() = self.name_cloned().or_else(|| {
+                self.info()
+                    .use_autoname()
+                    .then(|| self.get_autoname().ok())
+                    .flatten()
+                    .flatten()
+            })
         }
         hasher.update(&[self.name().is_some() as u8]);
         hasher.update(self.name().unwrap_or("").as_bytes());
