@@ -17,9 +17,7 @@ use uuid::Uuid;
 
 use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
-    circuit_node_private::{
-        CircuitNodeComputeInfoImpl, CircuitNodeHashItems, CircuitNodeMaybeAutoName,
-    },
+    circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
     new_rc,
     prelude::*,
     CachedCircuitInfo, CircuitFlags, PyCircuitBase, Shape, TensorEvalError,
@@ -27,7 +25,7 @@ use crate::{
 
 macro_rules! circuit_node_auto_leaf_impl {
     ($uuid:literal) => {
-        circuit_node_auto_impl!($uuid);
+        circuit_node_auto_impl!($uuid, no_autoname);
 
         fn children(&self) -> Box<dyn Iterator<Item = CircuitRc> + '_> {
             Box::new([].into_iter())
@@ -46,10 +44,6 @@ macro_rules! circuit_node_auto_leaf_impl {
     };
 }
 
-impl CircuitNodeMaybeAutoName for Array {}
-impl CircuitNodeMaybeAutoName for Scalar {}
-impl CircuitNodeMaybeAutoName for Symbol {}
-
 #[pyclass(extends=PyCircuitBase, unsendable)]
 #[derive(Clone)]
 pub struct Array {
@@ -227,9 +221,7 @@ impl CircuitNodeComputeInfoImpl for Symbol {
         self.info.shape.clone() // note: assumes shape has already been initted!
     }
     fn compute_flags(&self) -> CircuitFlags {
-        self.compute_flags_default()
-            & !CircuitFlags::NO_UNBOUND_SYMBOLS
-            & !CircuitFlags::IS_EXPLICITLY_COMPUTABLE
+        self.compute_flags_default() & !CircuitFlags::IS_EXPLICITLY_COMPUTABLE
     }
 }
 
