@@ -325,8 +325,10 @@ impl ModuleSpec {
         }
         recur_check_no_free_bound_by_nested(
             self.circuit.clone(),
-            &self.arg_specs.iter().map(|x| &x.symbol).collect::<Vec<_>>(),
-            &nodes.iter().collect::<Vec<_>>(),
+            SymsNodesInfo::new(
+                &self.arg_specs.iter().map(|x| &x.symbol).collect::<Vec<_>>(),
+                &nodes.iter().collect::<Vec<_>>(),
+            ),
             &mut Default::default(),
         )?;
         NO_FREE_BOUND_BY_NESTED.with(|cache| cache.borrow_mut().insert(key));
@@ -1371,65 +1373,92 @@ pub fn is_intersecting_free_syms(circuit: &CircuitRc, syms: &IndexSet<Symbol>) -
     }
 }
 
-pub fn recur_check_no_free_bound_by_nested(
+#[derive(Copy, Debug, Clone)]
+struct SymsNodesInfo<'a> {
+    syms: &'a [&'a Symbol],
+    nodes: &'a [&'a CircuitRc],
+    hash: HashBytes,
+}
+
+impl<'a> SymsNodesInfo<'a> {
+    pub fn new(syms: &'a [&'a Symbol], nodes: &'a [&'a CircuitRc]) -> Self {
+        let mut hasher = blake3::Hasher::new();
+
+        for sym in syms {
+            hasher.update(&sym.info().hash);
+        }
+        for n in nodes {
+            hasher.update(&n.info().hash);
+        }
+
+        Self {
+            syms,
+            nodes,
+            hash: hasher.finalize().into(),
+        }
+    }
+}
+
+fn recur_check_no_free_bound_by_nested(
     circ: CircuitRc,
-    syms: &[&Symbol],
-    nodes: &[&CircuitRc],
-    seen: &mut HashSet<HashBytes>,
+    info: SymsNodesInfo<'_>,
+    seen: &mut HashSet<(HashBytes, HashBytes)>,
 ) -> Result<()> {
-    if seen.insert(circ.info().hash) {
-        if let Some(m) = circ.as_module() {
-            // first check if we can term early (maybe not actually optimization)
-            if !is_intersecting_free_syms_iter(&circ, syms.iter().map(|x| *x)) {
-                return Ok(());
-            }
+    if !seen.insert((circ.info().hash, info.hash)) {
+        return Ok(());
+    }
 
-            // filter out bound symbols for recurring into
-            let (syms, nodes): (Vec<_>, Vec<&CircuitRc>) = syms
-                .iter()
-                .zip(nodes)
-                .filter_map(|(s, n)| {
-                    for arg_spec in &m.spec.arg_specs {
-                        if &arg_spec.symbol == *s {
-                            return None;
-                        }
-                    }
-                    Some((s, n))
-                })
-                .unzip();
+    if let Some(m) = circ.as_module() {
+        // filter out bound symbols for recurring into
+        let (syms, nodes): (Vec<_>, Vec<&CircuitRc>) = info
+            .syms
+            .iter()
+            .zip(info.nodes)
+            .filter(|(s, _)| {
+                // if the module binds the sym or doesn't have this symbol free, filter out
+                m.spec
+                    .arg_specs
+                    .iter()
+                    .all(|arg_spec| &arg_spec.symbol != **s)
+                    && has_free_sym(&m.spec.circuit, *s)
+            })
+            .unzip();
 
-            if is_intersecting_free_syms_iter(&m.spec.circuit, syms.iter().map(|x| *x)) {
-                // fail earlier if possible
-                recur_check_no_free_bound_by_nested(m.spec.circuit.clone(), &syms, &nodes, seen)?;
+        if !syms.is_empty() {
+            // fail earlier if possible
+            recur_check_no_free_bound_by_nested(
+                m.spec.circuit.clone(),
+                SymsNodesInfo::new(&syms, &nodes),
+                seen,
+            )?;
 
-                // then check valid
-                for (&sub_for_sym, &sym) in nodes.iter().zip(&syms) {
-                    if m.spec
-                        .arg_specs
-                        .iter()
-                        .any(|x| has_free_sym(sub_for_sym, &x.symbol))
-                    {
-                        bail!(SubstitutionError::CircuitHasFreeSymsBoundByNestedModule {
-                            circ: sub_for_sym.clone(),
-                            sym: sym.clone(),
-                            nested_module: m.clone(),
-                            bound_by_nested: m
-                                .spec
-                                .arg_specs
-                                .iter()
-                                .filter_map(|arg_spec| has_free_sym(sub_for_sym, &arg_spec.symbol)
-                                    .then(|| arg_spec.symbol.clone()))
-                                .collect()
-                        })
-                    }
+            // then check valid
+            for (&sub_for_sym, &sym) in nodes.iter().zip(&syms) {
+                if m.spec
+                    .arg_specs
+                    .iter()
+                    .any(|x| has_free_sym(sub_for_sym, &x.symbol))
+                {
+                    bail!(SubstitutionError::CircuitHasFreeSymsBoundByNestedModule {
+                        circ: sub_for_sym.clone(),
+                        sym: sym.clone(),
+                        nested_module: m.clone(),
+                        bound_by_nested: m
+                            .spec
+                            .arg_specs
+                            .iter()
+                            .filter_map(|arg_spec| has_free_sym(sub_for_sym, &arg_spec.symbol)
+                                .then(|| arg_spec.symbol.clone()))
+                            .collect()
+                    })
                 }
             }
         }
-
-        circ.non_free_children()
-            .map(|child| recur_check_no_free_bound_by_nested(child, syms, nodes, seen))
-            .collect::<Result<Vec<_>>>()?;
     }
+
+    circ.non_free_children()
+        .map(|child| recur_check_no_free_bound_by_nested(child, info, seen))
+        .collect::<Result<Vec<_>>>()?;
     Ok(())
 }
 
