@@ -173,9 +173,9 @@ pub fn generalfunction_pull_concat(generalfunction: &GeneralFunction) -> Option<
                     let span_here = cnode.info().shape[concat.axis];
                     let new_operands = zip(
                         &generalfunction.nodes,
-                        generalfunction.input_batchability,
+                        &generalfunction.input_batchability,
                     )
-                    .map(|(node, batchability)| {
+                    .map(|(node, &batchability)| {
                         if !batchability {
                             node.clone()
                         } else if node == operand {
