@@ -1,21 +1,24 @@
-use std::cmp::max;
+use std::{
+    cmp::max,
+    sync::{atomic::AtomicBool, mpsc::channel, Arc},
+};
 
 use anyhow::{bail, Result};
 use itertools::Itertools;
-use pyo3::prelude::*;
+use once_cell::unsync::Lazy;
+use rand::{rngs::SmallRng, Rng, SeedableRng};
 use rr_util::{
-    python_println, sv, timed,
+    python_println, sv, timed, timed_value,
     util::{is_unique, HashBytes},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
 
 use crate::{scheduled_execution::SchedulingOOMError, scheduling_z3::schedule_dag_strategy_ints};
+#[derive(Clone, Debug)]
 pub struct DagSimpSettings {
     pub same_ratio: f64,
     pub tiny_size: usize,
-    pub max_exhaustive_size: usize,
-    pub max_z3_ints_size: usize,
     pub acceptable_subdag_max: usize,
     pub exhaustive_settle_ns: usize,
     pub exhaustive_give_up_ns: usize,
@@ -23,6 +26,7 @@ pub struct DagSimpSettings {
     pub num_mem_chunks: usize,
     pub timeout: usize,
     pub verbose: usize,
+    pub parallelism: Option<usize>,
 }
 
 impl Default for DagSimpSettings {
@@ -30,8 +34,6 @@ impl Default for DagSimpSettings {
         Self {
             same_ratio: 0.25,
             tiny_size: 200_000,
-            max_exhaustive_size: 60,
-            max_z3_ints_size: 4000,
             acceptable_subdag_max: 100_000_000,
             exhaustive_settle_ns: 2_000_000,
             exhaustive_give_up_ns: 400_000_000,
@@ -39,12 +41,12 @@ impl Default for DagSimpSettings {
             timeout: 5_000, // useless defaults, todo fix
             num_mem_chunks: 200,
             mem_limit: 10,
+            parallelism: Some(8),
         }
     }
 }
 
 // rule: there can be only one edge between each pair of nodes
-#[pyclass]
 #[derive(Default, Clone, Debug)]
 pub struct Dag {
     pub children: Vec<Sv<[u32; 6]>>,
@@ -156,11 +158,6 @@ impl Dag {
     // simplify all
 
     pub fn simplify(&mut self, simp_settings: &DagSimpSettings) -> Result<bool> {
-        // println!(
-        //     "simplifying to_orig {:?} children {:?}",
-        //     &self.node_to_orig, &self.children
-        // );
-        // self.print_cytoscape_link();
         let mut result = false;
         while self.simplify_pass(simp_settings, false)? {
             if simp_settings.verbose > 2 {
@@ -179,7 +176,6 @@ impl Dag {
         if self.cost() > simp_settings.mem_limit {
             bail!(SchedulingOOMError::Simp {
                 cyto: self.repr_cytoscape_link(),
-                origin: "post check".to_owned()
             })
         }
         Ok(result)
@@ -625,49 +621,30 @@ impl Dag {
                     .unwrap()
                     .0,
             );
-            let order: Vec<(u32, usize)> = if iso.2.len() < settings.max_exhaustive_size
-                || iso.2.len() > settings.max_exhaustive_size
+            let exhaustive_scheduler = ExhaustiveScheduler::new(
+                direct_only_sub.clone(),
+                required_first,
+                settings.clone(),
+                None,
+            );
+            let resy = if direct_only_sub.node_to_orig.len() > 120 && settings.parallelism.is_some()
             {
-                let resy = timed!(
-                    direct_only_sub.schedule_exhaustive(required_first, settings),
+                timed!(
+                    exhaustive_scheduler.run_portfolio_release_gil(),
                     10,
                     settings.verbose >= 2
-                );
-                if resy.is_err() && did_anything {
-                    return Ok(true);
-                }
-
-                resy?
-                    .iter()
-                    .map(|(k, v)| (direct_only_sub.node_to_orig[k][0], *v))
-                    .collect()
+                )
             } else {
-                let order = direct_only_sub
-                    .compute_schedule_top_ints(
-                        settings.verbose,
-                        settings.mem_limit,
-                        settings.num_mem_chunks,
-                        settings.timeout,
-                        required_first.map(|x| x as usize),
-                    )
-                    .unwrap();
-                let inner_and_costs = direct_only_sub.get_costs_of_order(
-                    &order, // bad unwrap todo fix
-                );
-                let result: Vec<(u32, usize)> = inner_and_costs
-                    .into_iter()
-                    .map(|(n, c)| (direct_only_sub.node_to_orig[&n][0], c))
-                    .collect();
-                if result.len() != direct_only_sub.node_costs.len() || !is_unique(&order) {
-                    panic!(
-                        "order wrong length {:?} {:?}, {}",
-                        &order,
-                        &result,
-                        direct_only_sub.node_costs.len()
-                    )
-                }
-                result
+                timed!(exhaustive_scheduler.run(None), 10, settings.verbose >= 2)
             };
+            if resy.is_err() && did_anything {
+                return Ok(true);
+            }
+
+            let order = resy?
+                .iter()
+                .map(|(k, v)| (direct_only_sub.node_to_orig[k][0], *v))
+                .collect();
             // assert!(order[0].0==iso.1 && order.last().unwrap().0==iso.0);
             self.replace_isolated_with_linear(&order);
             if settings.verbose > 2 {
@@ -725,118 +702,6 @@ impl Dag {
         result
     }
 
-    pub fn schedule_exhaustive(
-        &self,
-        entry_point: Option<u32>,
-        settings: &DagSimpSettings,
-    ) -> Result<Vec<(u32, usize)>> {
-        let mut parents_left: Vec<i32> = self.parents.iter().map(|x| x.len() as i32).collect();
-        let mut children_left: Vec<i32> = self.children.iter().map(|x| x.len() as i32).collect();
-        let mut best: (Vec<(u32, usize)>, usize) = (vec![], settings.acceptable_subdag_max + 1);
-        let mut order: Vec<(u32, u32, usize)> = Vec::with_capacity(self.node_costs.len());
-
-        let mut options_stack: Vec<Sv<[u32; 16]>> = Vec::with_capacity(self.node_costs.len() + 2);
-        options_stack.push(
-            entry_point
-                .map(|z| sv![z])
-                .unwrap_or(filter_to_idx(&children_left, |x| *x == 0).collect()),
-        );
-        // want functions that modify stuff, using macros :|
-        macro_rules! choose {
-            ($i:expr) => {{
-                let options = options_stack.last().unwrap();
-                let choice = options[$i as usize] as u32;
-                for child in &self.children[choice as usize] {
-                    parents_left[*child as usize] -= 1;
-                }
-                for parent in &self.parents[choice as usize] {
-                    children_left[*parent as usize] -= 1;
-                }
-                children_left[choice as usize] -= 1;
-                let cost: usize = (0..parents_left.len() as u32)
-                    .filter(|i| children_left[*i as usize] < 0 && parents_left[*i as usize] >= 1)
-                    .map(|x| self.node_costs[x as usize])
-                    .sum();
-                order.push((choice, $i, cost));
-                options_stack.push(filter_to_idx(&children_left, |x| *x == 0).collect());
-                // trying out sorting options, must do this before stack is used
-                let endel1 = options_stack.len() - 1;
-                options_stack[endel1].sort_by_cached_key(|&x| {
-                    self.parents[x as usize]
-                        .iter()
-                        .map(|&parent| children_left[parent as usize])
-                        .max()
-                        .unwrap_or(0)
-                })
-            }};
-        }
-        macro_rules! pop {
-            () => {{
-                let popped = order.pop().unwrap();
-                options_stack.pop();
-                let choice = popped.0;
-                // println!("choice {}", choice);
-                for child in &self.children[choice as usize] {
-                    parents_left[*child as usize] += 1;
-                }
-                for parent in &self.parents[choice as usize] {
-                    children_left[*parent as usize] += 1;
-                }
-                children_left[choice as usize] += 1;
-                // println!("popped");
-                // dbg!(&order, &options_stack, &parents_left, &children_left, &live);
-
-                popped.1
-            }};
-        }
-        macro_rules! backtrack {
-            () => {{
-                while !order.is_empty()
-                    && order.last().unwrap().1 + 1 >= options_stack[order.len() - 1].len() as u32
-                {
-                    pop!();
-                }
-                if !order.is_empty() {
-                    let popped_idx = pop!();
-                    if !order.is_empty() {
-                        choose!(popped_idx + 1);
-                    }
-                }
-            }};
-        }
-        let start_instant = std::time::Instant::now();
-        for numiters in 0usize.. {
-            // dbg!(&order, &options_stack, &parents_left, &children_left, &live);
-            if order.len() == self.node_costs.len() {
-                best = (
-                    order.iter().map(|(a, _b, c)| (*a, *c)).collect(),
-                    *order.iter().map(|(_n, _i, c)| c).max().unwrap(),
-                );
-                if start_instant.elapsed().as_nanos() > settings.exhaustive_settle_ns as u128 {
-                    break;
-                }
-                backtrack!();
-            } else {
-                choose!(0);
-                if order.last().unwrap().2 >= best.1 {
-                    backtrack!();
-                }
-            }
-            if order.len() == 0 {
-                break;
-            }
-            if numiters.trailing_zeros() > 8
-                && start_instant.elapsed().as_nanos() > settings.exhaustive_give_up_ns as u128
-            {
-                bail!(SchedulingOOMError::Simp {
-                    cyto: self.repr_cytoscape_link(),
-                    origin: "exhaustive timeout".to_owned()
-                })
-            }
-        }
-        Ok(best.0)
-    }
-
     pub fn replace_isolated_with_linear(&mut self, order: &Vec<(u32, usize)>) {
         // println!("linearizing order {:?}", order);
         // because this is isolated except for first children and last parents,
@@ -1007,3 +872,204 @@ where
         .filter(move |(_i, x)| f(x))
         .map(|(i, _x)| i as u32)
 }
+#[derive(Clone, Debug)]
+struct ExhaustiveScheduler {
+    dag: Dag,
+    parents_left: Vec<i32>,
+    children_left: Vec<i32>,
+    best: (Vec<(u32, usize)>, usize),
+    order: Vec<(u32, u32, usize)>,
+    options_stack: Vec<Sv<[u32; 16]>>,
+    settings: DagSimpSettings,
+    found_anything: bool,
+    rng: Option<SmallRng>,
+}
+
+impl ExhaustiveScheduler {
+    pub fn new(
+        dag: Dag,
+        entry_point: Option<u32>,
+        settings: DagSimpSettings,
+        random_seed: Option<u64>,
+    ) -> Self {
+        let mut result = Self {
+            parents_left: dag.parents.iter().map(|x| x.len() as i32).collect(),
+            children_left: dag.children.iter().map(|x| x.len() as i32).collect(),
+            best: (vec![], settings.acceptable_subdag_max + 1),
+            order: Vec::with_capacity(dag.node_costs.len()),
+            options_stack: Vec::with_capacity(dag.node_costs.len() + 2),
+            dag,
+            settings,
+            found_anything: false,
+            rng: random_seed.map(|rs| SmallRng::seed_from_u64(rs)),
+        };
+        result.options_stack.push(
+            entry_point
+                .map(|z| sv![z])
+                .unwrap_or(filter_to_idx(&result.children_left, |x| *x == 0).collect()),
+        );
+        result
+    }
+
+    #[inline]
+    fn choose(&mut self, i: u32) {
+        let options = self.options_stack.last().unwrap();
+        let choice = options[i as usize] as u32;
+        for child in &self.dag.children[choice as usize] {
+            self.parents_left[*child as usize] -= 1;
+        }
+        for parent in &self.dag.parents[choice as usize] {
+            self.children_left[*parent as usize] -= 1;
+        }
+        self.children_left[choice as usize] -= 1;
+        let cost: usize = (0..self.parents_left.len() as u32)
+            .filter(|i| self.children_left[*i as usize] < 0 && self.parents_left[*i as usize] >= 1)
+            .map(|x| self.dag.node_costs[x as usize])
+            .sum();
+        self.order.push((choice, i, cost));
+        self.options_stack
+            .push(filter_to_idx(&self.children_left, |x| *x == 0).collect());
+        let endel1 = self.options_stack.len() - 1;
+        if let Some(rng) = &mut self.rng {
+            let rand_byte: u8 = rng.gen_range(0..5);
+            if rand_byte == 0 {
+                self.options_stack[endel1].reverse();
+                return;
+            } else if rand_byte == 1 {
+                return;
+            } else if rand_byte == 2 {
+                self.options_stack[endel1]
+                    .sort_by_cached_key(|&x| self.dag.children[x as usize].len());
+                return;
+            } else if rand_byte == 3 {
+                self.options_stack[endel1].sort_by_cached_key(|&x| {
+                    self.dag.parents[x as usize]
+                        .iter()
+                        .map(|&parent| self.children_left[parent as usize])
+                        .max()
+                        .unwrap_or(0)
+                });
+                return;
+            }
+        }
+        // trying out sorting options, must do this before stack is used
+        self.options_stack[endel1].sort_by_cached_key(|&x| {
+            // lower means tried sooner
+            self.dag.parents[x as usize]
+                .iter()
+                .map(|&parent| self.children_left[parent as usize])
+                .max()
+                .unwrap_or(0)
+                + (self.dag.children[x as usize].is_empty() as i32) * 10
+        })
+    }
+    #[inline]
+    fn pop(&mut self) -> u32 {
+        let popped = self.order.pop().unwrap();
+        self.options_stack.pop();
+        let choice = popped.0;
+        for child in &self.dag.children[choice as usize] {
+            self.parents_left[*child as usize] += 1;
+        }
+        for parent in &self.dag.parents[choice as usize] {
+            self.children_left[*parent as usize] += 1;
+        }
+        self.children_left[choice as usize] += 1;
+
+        popped.1
+    }
+    #[inline]
+    fn backtrack(&mut self) {
+        while !self.order.is_empty()
+            && self.order.last().unwrap().1 + 1
+                >= self.options_stack[self.order.len() - 1].len() as u32
+        {
+            self.pop();
+        }
+        if !self.order.is_empty() {
+            let popped_idx = self.pop();
+            if !self.order.is_empty() {
+                self.choose(popped_idx + 1);
+            }
+        }
+    }
+    fn run(mut self, halt_signal: Option<Arc<AtomicBool>>) -> Result<Vec<(u32, usize)>> {
+        let start_instant = std::time::Instant::now();
+        let mut numiters: usize = 0;
+        loop {
+            // println!("{}", order.iter().map(|x| x.0.to_string()).join(","));
+            if self.order.len() == self.dag.node_costs.len() {
+                self.best = (
+                    self.order.iter().map(|(a, _b, c)| (*a, *c)).collect(),
+                    *self.order.iter().map(|(_n, _i, c)| c).max().unwrap(),
+                );
+                self.found_anything = true;
+                if start_instant.elapsed().as_nanos() > self.settings.exhaustive_settle_ns as u128 {
+                    break;
+                }
+                self.backtrack();
+            } else {
+                self.choose(0);
+                if self.order.last().unwrap().2 >= self.best.1 {
+                    self.backtrack();
+                }
+            }
+            if self.order.len() == 0
+                || numiters.trailing_zeros() > 8
+                    && (start_instant.elapsed().as_nanos()
+                        > self.settings.exhaustive_give_up_ns as u128
+                        || halt_signal
+                            .as_ref()
+                            .map(|h| h.load(std::sync::atomic::Ordering::Relaxed))
+                            .unwrap_or(false))
+            {
+                break;
+            }
+            numiters += 1;
+        }
+        if self.found_anything {
+            return Ok(self.best.0);
+        }
+        bail!(SchedulingOOMError::ExhaustiveTimeout {
+            cyto: self.dag.repr_cytoscape_link(),
+            iters: numiters,
+            size: self.dag.node_to_orig.len(),
+        })
+        // self.print_cytoscape_link();
+    }
+
+    pub fn run_portfolio_release_gil(self) -> Result<Vec<(u32, usize)>> {
+        pyo3::Python::with_gil(|py| py.allow_threads(move || self.run_portfolio_raw()))
+    }
+
+    pub fn run_portfolio_raw(self) -> Result<Vec<(u32, usize)>> {
+        // println!("portfolio!");
+        let halt_signal: Arc<AtomicBool> = Arc::new(AtomicBool::from(false));
+        let (tx, rx) = channel();
+        for i in 0..self.settings.parallelism.unwrap() {
+            let tx = tx.clone();
+            let mut self_here = self.clone();
+            self_here.rng = if i != 0 {
+                None
+            } else {
+                Some(SmallRng::seed_from_u64(i as u64))
+            };
+            let halt_signal_clone = halt_signal.clone();
+            std::thread::spawn(move || {
+                // let randy = self_here.use_rand;
+                let (result, _time) = timed_value!(self_here.run(Some(halt_signal_clone)));
+                // println!("finished {} {} took {:.2?}", result.is_ok(), randy, time);
+                tx.send(result).unwrap_or(());
+            });
+        }
+        let mut resulty: Result<Vec<(u32, usize)>> = Err(SchedulingOOMError::ThreadsLost {}.into());
+        for recv in rx.iter().take(self.settings.parallelism.unwrap()) {
+            resulty = recv;
+            if resulty.is_ok() {
+                break;
+            }
+        }
+        halt_signal.swap(true, std::sync::atomic::Ordering::Relaxed);
+        return resulty;
+    }
+}
