@@ -3,7 +3,7 @@ use circuit_base::{CircuitNode, CircuitRc, Symbol};
 use circuit_rewrites::circuit_manipulation::replace_nodes_py;
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
-use rr_util::{cached_method, pycall, util::HashBytes};
+use rr_util::{cached_method, caching::FastUnboundedCache, pycall, util::HashBytes};
 use rustc_hash::FxHashMap as HashMap;
 
 use crate::{iterative_matcher::function_per_child_op, IterativeMatcherRc};
@@ -59,7 +59,7 @@ where
     // using caching struct so i can mutate back map while cached recursing
     struct Thing {
         back_map: HashMap<CircuitRc, CircuitRc>,
-        estimated_cache: cached::UnboundCache<(HashBytes, IterativeMatcherRc), CircuitRc>,
+        estimated_cache: FastUnboundedCache<(HashBytes, IterativeMatcherRc), CircuitRc>,
     }
     impl Thing {
         #[apply(cached_method)]
@@ -94,7 +94,7 @@ where
     }
     let mut thing = Thing {
         back_map: Default::default(),
-        estimated_cache: cached::UnboundCache::new(),
+        estimated_cache: FastUnboundedCache::default(),
     };
     let circ_result = thing.recurse(circuit, traversal, &namer)?;
     Ok((circ_result, thing.back_map))
