@@ -85,7 +85,7 @@ impl Display for IterativeMatcher {
                     indented_if_multiline(&format!("{}", pyobject))
                 )
             }
-            IterativeMatcherData::Filter(filter) => {
+            IterativeMatcherData::Restrict(filter) => {
                 let mut entries: Vec<String> = vec![];
                 if filter.start_depth.is_some() || filter.end_depth.is_some() {
                     entries.push(format!(
@@ -107,7 +107,7 @@ impl Display for IterativeMatcher {
                     entries.push(format!("cur_depth:{}", filter.depth));
                 }
                 format!(
-                    "Filter{{{}}}",
+                    "Restrict{{{}}}",
                     maybe_indented_list(entries, ", ", false, true)
                 )
             }
