@@ -16,7 +16,9 @@ use itertools::Itertools;
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
-    cached_method, fn_struct,
+    cached_method,
+    caching::FastUnboundedCache,
+    fn_struct,
     py_types::{scalar_to_tensor, PyShape, Tensor, PY_UTILS},
     python_error_exception, sv,
     tensor_util::{
@@ -25,7 +27,7 @@ use rr_util::{
     tu8v,
     util::{flip_op_result, HashBytes},
 };
-use rustc_hash::FxHashSet as HashSet;
+use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
 use thiserror::Error;
 
@@ -108,7 +110,7 @@ pub struct RandomSampleSpec {
     #[pyo3(get, set)]
     seeder: Option<Seeder>,
     // kinda dumb that we use a mutex here. Certainly overkill in a particular sense...
-    indices_cache: Arc<Mutex<cached::UnboundCache<HashBytes, Tensor>>>,
+    indices_cache: Arc<Mutex<FastUnboundedCache<HashBytes, Tensor>>>,
 }
 
 macro_rules! py_forward_sample_spec {
@@ -209,8 +211,8 @@ impl RandomSampleSpec {
         .into_init()
     }
 
-    pub fn indices_cache(&self) -> ::std::collections::HashMap<HashBytes, Tensor> {
-        self.indices_cache.lock().unwrap().get_store().clone()
+    pub fn indices_cache(&self) -> HashMap<HashBytes, Tensor> {
+        self.indices_cache.lock().unwrap().0.clone()
     }
 }
 
@@ -266,7 +268,7 @@ impl Default for RandomSampleSpec {
             replace: true,
             probs_and_group_evaluation_settings: Default::default(),
             seeder: None,
-            indices_cache: Arc::new(Mutex::new(cached::UnboundCache::new())),
+            indices_cache: Arc::new(Mutex::new(FastUnboundedCache::default())),
         }
     }
 }
@@ -275,17 +277,20 @@ fn combine_dtype_device_for_sample(
     var: &DiscreteVar,
     device_dtype: TorchDeviceDtypeOp,
 ) -> Result<TorchDeviceDtype> {
+    let var_probs_dtype_device = TorchDeviceDtypeOp {
+        device: var.info().device_dtype.device,
+        dtype: var.probs_and_group.info().device_dtype.dtype,
+    };
     device_dtype
         .clone()
-        .combine(var.info().device_dtype.clone())
+        .combine(var_probs_dtype_device)
         .with_context(|| {
             format!(
                 concat!(
                     "sampled var device_dtype failed to combine with passed device_dtype",
-                    "var device_dtype={:?} device_dtype={:?}"
+                    "var probs dtype_device={:?} passed device_dtype={:?}"
                 ),
-                &var.info().device_dtype,
-                device_dtype
+                var_probs_dtype_device, device_dtype
             )
         })
         .map(|x| x.unwrap_or_defaults())
@@ -559,7 +564,7 @@ pub struct Sampler {
     sample_spec: SampleSpec,
     #[pyo3(get)]
     run_on_sampled: TransformRc,
-    estimated_cache: cached::UnboundCache<HashBytes, CircuitRc>,
+    estimated_cache: FastUnboundedCache<HashBytes, CircuitRc>,
 }
 
 // TODO: add more configs/call backs as needed!
@@ -590,7 +595,7 @@ impl Sampler {
             expander,
             cumulant_matcher,
             sample_spec,
-            estimated_cache: cached::UnboundCache::new(),
+            estimated_cache: FastUnboundedCache::default(),
             run_on_sampled,
         }
     }
