@@ -1210,7 +1210,7 @@ impl ModuleConstructUpdateBindings {
             (m.spec.circuit.clone(), m.spec.circuit.clone(), Vec::new())
         };
 
-        let new_modules = if !self.is_flatten {
+        let new_modules = if self.is_flatten {
             applied_modules.to_vec()
         } else {
             let mut new_modules = prior_modules.clone();
@@ -1266,15 +1266,22 @@ impl ModuleConstructUpdateBindings {
     }
 }
 
-#[pyfunction(bind_name = "\"upd_bind\".to_owned()", short_if_not_leaf = "true")]
+#[pyfunction(
+    bind_name = "\"upd_bind\".to_owned()",
+    short_if_not_leaf = "true",
+    keep_name_if_not_leaf = "false"
+)]
 pub fn default_update_bindings_nested_namer(
     bind_name: String,
     short_if_not_leaf: bool,
+    keep_name_if_not_leaf: bool,
 ) -> NestedModuleNamer {
     let default_overall_namer = default_nested_module_namer(bind_name.clone());
     NestedModuleNamer::Dyn(NestedModuleNamerDynStruct(Arc::new(
         move |base_circuit, running_circuit, modules, pushed_overall_mod_count| {
-            if pushed_overall_mod_count.is_none() && short_if_not_leaf {
+            if pushed_overall_mod_count.is_none() && keep_name_if_not_leaf {
+                Ok(base_circuit.name_cloned())
+            } else if pushed_overall_mod_count.is_none() && short_if_not_leaf {
                 Ok(base_circuit
                     .name()
                     .map(|name| format!("{name} {bind_name}")))
@@ -1291,7 +1298,7 @@ pub fn default_update_bindings_nested_namer(
 }
 
 #[pyfunction(
-    namer = "default_update_bindings_nested_namer(\"upd_bind\".to_owned(), true)",
+    namer = "default_update_bindings_nested_namer(\"upd_bind\".to_owned(), true, false)",
     skip_module = "MatcherData::Always(false).into()",
     run_update_on_new_spec_circuits = "false",
     flatten_modules = "false"
